/*
 * This program has grown organically from being a proof-of-concept
 * at a beer-filled conference to a full-fledged multi-purpose
 * plugin. As such, it needs some cleanups...
 */

#define _GNU_SOURCE 1
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <stdarg.h>
#include <ctype.h>
#include <libgen.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include "check_snmpif.h"

#define ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))

#define STATE_OK 0
#define STATE_WARNING 1
#define STATE_CRITICAL 2
#define STATE_UNKNOWN 3

/* for ifOperStatus and ifAdminStatus */
#define IFSTATUS_UP 1
#define IFSTATUS_DOWN 2
#define IFSTATUS_TESTING 3
#define IFSTATUS_UNKNOWN 4
#define IFSTATUS_DORMANT 5
#define IFSTATUS_NOTPRESENT 6
#define IFSTATUS_LOWERLAYERDOWN 7

#define SNMPWALK_STOP 1

#define IFINFO_TRAFFIC_BPS_PERFDATA 1
#define IFINFO_TRAFFIC_NAGVIS_PERFDATA 2
#define IFINFO_ERRORS_IGNORE_DISCARDS 1

#define walker_callback(x) int (*x)(netsnmp_session *, netsnmp_variable_list *, void *, void *)

typedef unsigned long long u64;

static int ignore_ifx;
static int verbose;
static oid uptime_oid[MAX_OID_LEN];
static size_t uptime_oid_len = MAX_OID_LEN;

static unsigned long device_uptime; /* uptime of the targeted device */

static const char *humanize_bytes(double bytes) __attribute__((unused));
struct options {
	int dormant_state;
	int admin_down_state;
	int verbose;
	int flags;
	const char *path;
};

extern const char *__progname;

typedef struct if_info {
	/* book-keeping */
	unsigned int valid_vars;
	struct timeval tv;

	int Index;                                /* 1 */
	char *Descr;
	int Type;
	/* int Mtu; */
	unsigned long Speed;                      /* 5 */
	/* char *PhysAddress; */
	int AdminStatus;
	int OperStatus;
	/* timeticks *LastChange */
	unsigned long InOctets;                   /* 10 */
	/* int InUcastPkts; */
	/* int InNUcastPkts; */
	unsigned int InDiscards;
	unsigned int InErrors;
	/* int InUnknownProtos; */                /* 15 */
	unsigned long OutOctets;
	/* int OutUcastPkts; */
	/* int OutNUcastPkts; */
	unsigned int OutDiscards;
	unsigned int OutErrors;                   /* 20 */
	/* int OutQLen; */

	/* from subtree .1.3.6.1.2.1.31.1.1.1 */
	unsigned long HighSpeed;
	char *Name;
	unsigned long InUcastPkts;
	unsigned long InMulticastPkts;
	unsigned long InBroadcastPkts;
	unsigned long OutUcastPkts;
	unsigned long OutMulticastPkts;
	unsigned long OutBroadcastPkts;

#define HCOutOctets OutOctets
#define HCInOctets InOctets
	unsigned int vars;
	unsigned int xvars;
} if_info;

typedef struct traffic_info {
	double bits, bps, pct, ops;
	double ucastpps, bcastpps, mcastpps;
	unsigned long errors, discards;
	double eps, dps; /* errors per second / discards per second */
	long long octets;
	int use_pct;
} traf_info;


typedef struct threshold_info {
	traf_info in, out;
} threshold_info;

static struct interface_type {
	int code;
	const char *name;
} iftypes[] = {
	{ 0, "-- INVALID --" },
	{ 1, "other" },
	{ 2, "regular1822" },
	{ 3, "hdh1822" },
	{ 4, "ddnX25" },
	{ 5, "rfc877x25" },
	{ 6, "ethernetCsmacd" },
	{ 7, "iso88023Csmacd" },
	{ 8, "iso88024TokenBus" },
	{ 9, "iso88025TokenRing" },
	{ 10, "iso88026Man" },
	{ 11, "starLan" },
	{ 12, "proteon10Mbit" },
	{ 13, "proteon80Mbit" },
	{ 14, "hyperchannel" },
	{ 15, "fddi" },
	{ 16, "lapb" },
	{ 17, "sdlc" },
	{ 18, "ds1" },
	{ 19, "e1" },
	{ 20, "basicISDN" },
	{ 21, "primaryISDN" },
	{ 22, "propPointToPointSerial" },
	{ 23, "ppp" },
	{ 24, "softwareLoopback" },
	{ 25, "eon" },
	{ 26, "ethernet3Mbit" },
	{ 27, "nsip" },
	{ 28, "slip" },
	{ 29, "ultra" },
	{ 30, "ds3" },
	{ 31, "sip" },
	{ 32, "frameRelay" },
	{ 33, "rs232" },
	{ 34, "para" },
	{ 35, "arcnet" },
	{ 36, "arcnetPlus" },
	{ 37, "atm" },
	{ 38, "miox25" },
	{ 39, "sonet" },
	{ 40, "x25ple" },
	{ 41, "iso88022llc" },
	{ 42, "localTalk" },
	{ 43, "smdsDxi" },
	{ 44, "frameRelayService" },
	{ 45, "v35" },
	{ 46, "hssi" },
	{ 47, "hippi" },
	{ 48, "modem" },
	{ 49, "aal5" },
	{ 50, "sonetPath" },
	{ 51, "sonetVT" },
	{ 52, "smdsIcip" },
	{ 53, "propVirtual" },
	{ 54, "propMultiplexor" },
	{ 55, "ieee80212" },
	{ 56, "fibreChannel" },
	{ 57, "hippiInterface" },
	{ 58, "frameRelayInterconnect" },
	{ 59, "aflane8023" },
	{ 60, "aflane8025" },
	{ 61, "cctEmul" },
	{ 62, "fastEther" },
	{ 63, "isdn" },
	{ 64, "v11" },
	{ 65, "v36" },
	{ 66, "g703at64k" },
	{ 67, "g703at2mb" },
	{ 68, "qllc" },
	{ 69, "fastEtherFX" },
	{ 70, "channel" },
	{ 71, "ieee80211" },
	{ 72, "ibm370parChan" },
	{ 73, "escon" },
	{ 74, "dlsw" },
	{ 75, "isdns" },
	{ 76, "isdnu" },
	{ 77, "lapd" },
	{ 78, "ipSwitch" },
	{ 79, "rsrb" },
	{ 80, "atmLogical" },
	{ 81, "ds0" },
	{ 82, "ds0Bundle" },
	{ 83, "bsc" },
	{ 84, "async" },
	{ 85, "cnr" },
	{ 86, "iso88025Dtr" },
	{ 87, "eplrs" },
	{ 88, "arap" },
	{ 89, "propCnls" },
	{ 90, "hostPad" },
	{ 91, "termPad" },
	{ 92, "frameRelayMPI" },
	{ 93, "x213" },
	{ 94, "adsl" },
	{ 95, "radsl" },
	{ 96, "sdsl" },
	{ 97, "vdsl" },
	{ 98, "iso88025CRFPInt" },
	{ 99, "myrinet" },
	{ 100, "voiceEM" },
	{ 101, "voiceFXO" },
	{ 102, "voiceFXS" },
	{ 103, "voiceEncap" },
	{ 104, "voiceOverIp" },
	{ 105, "atmDxi" },
	{ 106, "atmFuni" },
	{ 107, "atmIma" },
	{ 108, "pppMultilinkBundle" },
	{ 109, "ipOverCdlc" },
	{ 110, "ipOverClaw" },
	{ 111, "stackToStack" },
	{ 112, "virtualIpAddress" },
	{ 113, "mpc" },
	{ 114, "ipOverAtm" },
	{ 115, "iso88025Fiber" },
	{ 116, "tdlc" },
	{ 117, "gigabitEthernet" },
	{ 118, "hdlc" },
	{ 119, "lapf" },
	{ 120, "v37" },
	{ 121, "x25mlp" },
	{ 122, "x25huntGroup" },
	{ 123, "transpHdlc" },
	{ 124, "interleave" },
	{ 125, "fast" },
	{ 126, "ip" },
	{ 127, "docsCableMaclayer" },
	{ 128, "docsCableDownstream" },
	{ 129, "docsCableUpstream" },
	{ 130, "a12MppSwitch" },
	{ 131, "tunnel" },
	{ 132, "coffee" },
	{ 133, "ces" },
	{ 134, "atmSubInterface" },
	{ 135, "l2vlan" },
	{ 136, "l3ipvlan" },
	{ 137, "l3ipxvlan" },
	{ 138, "digitalPowerline" },
	{ 139, "mediaMailOverIp" },
	{ 140, "dtm" },
	{ 141, "dcn" },
	{ 142, "ipForward" },
	{ 143, "msdsl" },
	{ 144, "ieee1394" },
	{ 145, "if-gsn" },
	{ 146, "dvbRccMacLayer" },
	{ 147, "dvbRccDownstream" },
	{ 148, "dvbRccUpstream" },
	{ 149, "atmVirtual" },
	{ 150, "mplsTunnel" },
	{ 151, "srp" },
	{ 152, "voiceOverAtm" },
	{ 153, "voiceOverFrameRelay" },
	{ 154, "idsl" },
	{ 155, "compositeLink" },
	{ 156, "ss7SigLink" },
	{ 157, "propWirelessP2P" },
	{ 158, "frForward" },
	{ 159, "rfc1483" },
	{ 160, "usb" },
	{ 161, "ieee8023adLag" },
	{ 162, "bgppolicyaccounting" },
	{ 163, "frf16MfrBundle" },
	{ 164, "h323Gatekeeper" },
	{ 165, "h323Proxy" },
	{ 166, "mpls" },
	{ 167, "mfSigLink" },
	{ 168, "hdsl2" },
	{ 169, "shdsl" },
	{ 170, "ds1FDL" },
	{ 171, "pos" },
	{ 172, "dvbAsiIn" },
	{ 173, "dvbAsiOut" },
	{ 174, "plc" },
	{ 175, "nfas" },
	{ 176, "tr008" },
	{ 177, "gr303RDT" },
	{ 178, "gr303IDT" },
	{ 179, "isup" },
	{ 180, "propDocsWirelessMaclayer" },
	{ 181, "propDocsWirelessDownstream" },
	{ 182, "propDocsWirelessUpstream" },
	{ 183, "hiperlan2" },
	{ 184, "propBWAp2Mp" },
	{ 185, "sonetOverheadChannel" },
	{ 186, "digitalWrapperOverheadChannel" },
	{ 187, "aal2" },
	{ 188, "radioMAC" },
	{ 189, "atmRadio" },
	{ 190, "imt" },
	{ 191, "mvl" },
	{ 192, "reachDSL" },
	{ 193, "frDlciEndPt" },
	{ 194, "atmVciEndPt" },
	{ 195, "opticalChannel" },
	{ 196, "opticalTransport" },
	{ 197, "propAtm" },
	{ 198, "voiceOverCable" },
	{ 199, "infiniband" },
	{ 200, "teLink" },
	{ 201, "q2931" },
	{ 202, "virtualTg" },
	{ 203, "sipTg" },
	{ 204, "sipSig" },
	{ 205, "docsCableUpstreamChannel" },
	{ 206, "econet" },
	{ 207, "pon155" },
	{ 208, "pon622" },
	{ 209, "bridge" },
	{ 210, "linegroup" },
	{ 211, "voiceEMFGD" },
	{ 212, "voiceFGDEANA" },
	{ 213, "voiceDID" },
	{ 214, "mpegTransport" },
	{ 215, "sixToFour" },
	{ 216, "gtp" },
	{ 217, "pdnEtherLoop1" },
	{ 218, "pdnEtherLoop2" },
	{ 219, "opticalChannelGroup" },
	{ 220, "homepna" },
	{ 221, "gfp" },
	{ 222, "ciscoISLvlan" },
	{ 223, "actelisMetaLOOP" },
	{ 224, "fcipLink" },
	{ 225, "rpr" },
	{ 226, "qam" },
	{ 227, "lmp" },
	{ 228, "cblVectaStar" },
	{ 229, "docsCableMCmtsDownstream" },
	{ 230, "adsl2" },
	{ 231, "macSecControlledIF" },
	{ 232, "macSecUncontrolledIF" },
	{ 233, "aviciOpticalEther" },
	{ 234, "atmbond" },
	{ 235, "voiceFGDOS" },
	{ 236, "mocaVersion1" },
	{ 237, "ieee80216WMAN" },
	{ 238, "adsl2plus" },
	{ 239, "dvbRcsMacLayer" },
	{ 240, "dvbTdm" },
	{ 241, "dvbRcsTdma" },
	{ 242, "x86Laps" },
	{ 243, "wwanPP" },
	{ 244, "wwanPP2" },
	{ 245, "voiceEBS" },
	{ 246, "ifPwType" },
	{ 247, "ilan" },
	{ 248, "pip" },
	{ 249, "aluELP" },
	{ 250, "gpon" },
	{ 251, "vdsl2" },
};

static void debug(const char *fmt, ...)
{
	va_list ap;
	if (!verbose)
		return;
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static const char *iftype_str(unsigned int type)
{
	if (type >= ARRAY_SIZE(iftypes))
		return "OUT OF BOUNDS";
	return iftypes[type].name;
}

static const char *state_str(unsigned int state)
{
	static const char *state_strings[] = { "OK", "WARNING", "CRITICAL" };
	if (state < ARRAY_SIZE(state_strings)) {
		return state_strings[state];
	}
	return "UNKNOWN";
}

static void die(int status, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	exit(status);
}

static void usage(const char *fmt, ...)
{
	if (fmt) {
		va_list ap;
		va_start(ap, fmt);
		vprintf(fmt, ap);
		va_end(ap);
	}

	printf("Usage: %s <mode> -H <host> -<i|k|d|n> <interface> [options...]\n", __progname);
	printf("\nWhere 'mode' must be 'traffic', 'errors' or 'status'.\n\n");
	printf("Options (* = required; # = changeably required):\n");
	printf("   --debug                Enable (lots) of debugging output\n");
	printf(" * -H --host <host>       Hostname or ip to query\n");
	printf("   -C --community <str>   SNMP community. Defaults to public\n");
	printf(" # -i --index <int>       Interface index\n");
	printf(" # -k --key <int>         Interface index\n");
	printf(" # -d --descr <str>       Interface description (.1.3.6.1.2.1.2.2.1.2.*)\n");
	printf(" # -n --if-name <str>     Interface name (.1.3.6.1.2.1.31.1.1.1.1.*)\n");
	printf("   -v --version <1|2c|3>  SNMP protocol version (default: 1)\n");
	printf("   -t <int>               Timeout in seconds (default: 15)\n");
	printf("   -r <int>               Number of retries (default: 5)\n");
	printf("\n");
	printf("SNMP v3 options:\n");
	printf("   -L --seclevel <str>    'noAuthNoPriv', 'authNoPriv' or 'authPriv'\n");
	printf("   -U --secname <str>     username for SNMPv3 context\n");
	printf("   -c --context <str>     SNMPv3 context name (default is empty string)\n");
	printf("   -a --authprot <str>    Authentication protocol (md5 or sha1)\n");
	printf("   -A --authpass <str>    Authentication password (cleartext ascii or\n");
	printf("                            localized key in hex with 0x prefix generated\n");
	printf("                            by using the 'snmpkey' utility)\n");
	printf("                            Auth password and authEngineID\n");
	printf("   -X --privpass <str>    Privacy password (same rules as for authpass)\n");
	printf("   -P --privproto <str>   Privacy protocol (DES or AES; default: DES)\n");
	printf("   -p --port <int>        SNMP port (default: 161)\n");
	printf("   -h --help              This cruft\n");
	printf("   --hide-community|-hc   Noop option for backwards compatibility\n");
	printf("\nMode 'status' options:\n");
	printf("   -w --warn (i|w|c)      Result if interface is dormant (default=critical)\n");
	printf("   -D --admin-down        Result if interface adminstatus is DOWN (default=OK)\n");
	printf("\nMode 'errors' options:\n");
	printf("   -w --warn <float>      Errors+Discards/second resulting in warning\n");
	printf("      --warn-in <float>   Same as '--warn', but for inbound traffic only\n");
	printf("      --warn-out <float>  same as '--warn', but for outbound traffic only\n");
	printf("   -c --crit <float>      Errors+Discards/second resulting in critical\n");
	printf("      --crit-in <float>   Same as '--crit', but for inbound traffic only\n");
	printf("      --crit-out <float>  same as '--crit', but for outbound traffic only\n");
	printf("   -O --only-errors       Count only errors and ignore discards\n");
	printf("   -S --sampling-interval <int>  Use two samplings of traffic data instead\n");
	printf("                                   of relying on disc-stored data\n");
	printf("\nMode 'traffic' options:\n");
	printf("   -w --warn <int or pct> Bits per second OR percent of max capacity\n");
	printf("                            to trigger a warning result\n");
	printf("      --warn-in <float>   Same as '--warn', but for inbound traffic only\n");
	printf("      --warn-out <float>  same as '--warn', but for outbound traffic only\n");
	printf("   -c --crit <int or pct> Bits per second OR percent of max capacity\n");
	printf("                            to trigger a critical result\n");
	printf("      --crit-in <float>   Same as '--crit', but for inbound traffic only\n");
	printf("      --crit-out <float>  same as '--crit', but for outbound traffic only\n");
	printf("   -B --perfdatabits      Print bps in performance data rather than percent\n");
	printf("   -N --nagvis-perfdata   Print performance data for NagVis weathermap\n");
	printf("   -b --bps <int>         Interface traffic capacity in bits.\n");
	printf("                            Example: \"--bps 100m\" for a 100mbps interface\n");
	printf("   -S --sampling-interval <int>  Use two samplings of traffic data instead\n");
	printf("                                   of relying on disc-stored data\n");
	if (fmt)
		exit(STATE_UNKNOWN);
	exit(0);
}

static const char *interface_name(if_info *ii)
{
	static char name[256];

	if (ii->Name) {
		return ii->Name;
	}
	if (ii->Descr) {
		return ii->Descr;
	}
	snprintf(name, sizeof(name) - 1, "if%d", ii->Index);
	return name;
}

static const char *humanize_bits(double bps)
{
	static char buf[4][64];
	static int buf_i = 0;
	int ret = 0;
	char *p;
	const char *suffix[] = { "bps", "kbps", "Mbps", "Gbps", "Tbps" };
	unsigned int suff_i = 0, use_decimals = 0;

	while (suff_i < ARRAY_SIZE(suffix) && bps >= 1000) {
		bps /= 1000;
		suff_i++;
	}

	/* see if we should use int notation or floating point */
	use_decimals = bps != (unsigned long)bps;

	p = buf[buf_i++];
	if (use_decimals)
		ret = snprintf(p, sizeof(buf[0]), "%.2lf%s", bps, suffix[suff_i]);
	else
		ret = snprintf(p, sizeof(buf[0]), "%lu%s", (unsigned long)bps, suffix[suff_i]);

	if (ret > sizeof(buf[0])) {
		debug("Humanized bitrate string (of size %d) did not fit into buffer (of size %d)\n", ret, sizeof(buf[0]));
	}
	return p;
}

static const char *humanize_bytes(double bytes)
{
	static char buf[32][16];
	static int buf_i = 0;
	char *p;
	const char *suffix[] = { "byte", "KiB", "MiB", "GiB", "TiB", "PiB" };
	unsigned int suff_i = 0;

	while (suff_i < ARRAY_SIZE(suffix) && bytes > 1024) {
		bytes /= 1024;
		suff_i++;
	}
	p = buf[buf_i++];
	sprintf(p, "%.2lf%s", bytes, suffix[suff_i]);
	return p;
}

const char *get_db_path(char *dir, char *host, char *mode, if_info *ii)
{
	static char path[PATH_MAX];

	if (!mode && !ii) {
		/* IFINFO_ALL hack */
		snprintf(path, sizeof(path) - 1, "%s/combined/%s", dir, host);
		return path;
	}
	snprintf(path, sizeof(path) - 1, "%s/%s/if%d@%s.%s",
			 dir, mode, ii->Index, host, mode);
	return path;
}

static int db_load(const char *path, void *buf, unsigned int bufsize)
{
	int fd, len;
	struct stat st;

	if (stat(path, &st) < 0) {
		return -errno;
	}

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return -errno;

	if (st.st_mtime == time(NULL)) {
		/* this should be handled somehow */
	}

	if (st.st_size != bufsize) {
		/* this marks "incompatible" */
		close(fd);
		return -EL2NSYNC;
	}

	len = read(fd, buf, bufsize);
	close(fd);

	if (len != (int)bufsize)
		return -EL2NSYNC;

	return 0;
}

static int mkdir_p(const char *orig_path, mode_t mode)
{
	char *sep, *path;
	int ret = 0;

	if (!orig_path) {
		errno = EFAULT;
		return -1;
	}

	sep = path = strdup(orig_path);
	if (!sep)
		return -1;

	for (;;) {
		if ((sep = strchr(sep + 1, '/'))) {
			*sep = 0; /* nul-terminate path */
		} else {
			/* we always get full filepaths, so skip last entry */
			break;
		}

		debug("Trying mkdir(%s, ...)\n", path);
		/* try to mkdir(), and break on non-dir entries */
		if ((ret = mkdir(path, mode)) < 0) {
			struct stat st;
			if (errno != EEXIST) {
				debug("Got error at '%s': %m\n", path);
				break;
			}

			if (stat(path, &st) < 0) {
				debug("  stat() failed: %m\n");
				break;
			}
			if (!S_ISDIR(st.st_mode)) {
				debug("  not a directory. breaking\n");
				break;
			}
		}

		/* end of path or trailing slash? */
		if (!sep || !sep[1])
			break;
		*sep = '/';
	}

	if (ret) {
		debug("mkdir_p() breaking out at %s: %m\n", path);
	}

	free(path);
	return ret;
}

static int db_save(const char *path, void *buf, unsigned int buflen)
{
	int old_mask, fd, len;

	old_mask = umask(022);
	fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	umask(old_mask);

	if (fd < 0 && errno == ENOENT) {
		if (mkdir_p(path, 0755) < 0) {
			printf("Failed to create directory for %s: %m", path);
			return -errno;
		}
		old_mask = umask(022);
		fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0644);
		umask(old_mask);
	}
	if (fd < 0) {
		printf("failed to open %s: %s\n", path, strerror(errno));
		return -errno;
	}

	len = write(fd, buf, buflen);
	close(fd);
	if (len != (int)buflen) {
		return -EL2NSYNC;
	}

	return 0;
}

static int load_if_data(const char *path, if_info *ii)
{
	int ret;
	ret = db_load(path, ii, sizeof(*ii));
	ii->Name = ii->Descr = NULL;
	return ret;
}

int save_if_data(const char *path, if_info *ii)
{
	return db_save(path, ii, sizeof(*ii));
}

static u64 compute_ifspeed(if_info *ii)
{
	debug("Computing ifspeed. ii->Speed=%lu; ii->HighSpeed=%lu\n",
		ii->Speed, ii->HighSpeed);
	/*
	 * ii->HighSpeed is inaccurate for really low speeds, so
	 * prefer Speed if it's available and low enough
	 */
	if (ii->Speed && (signed)ii->Speed > 0) {
		debug("Using ii->Speed\n");
		return ii->Speed;
	}

	if (ii->HighSpeed) {
		debug("Using HighSpeed * 1000000\n");
		return ii->HighSpeed * 1000000;
	}

	/*
	 * ii->Speed only goes to (2^^32) - 1. If that's reached,
	 * we must assume that it's higher, so go with 10gbps
	 */
	if (ii->Speed == ~0U) {
		debug("Guessing at 10gbps\n");
		return 10 * 10000000000;
	}

	debug("Guessing at 100mbps\n");
	/* nothing's set, so take a wild guess at 100mbps */
	return 100000000;
}

static const char *oid2str(oid *o, size_t len)
{
	static char str[2 + (MAX_OID_LEN * 4)];
	size_t i = 0, x = 0;

	for (i = 0; i < len; i++) {
		str[x++] = '.';
		if (o[i] >= 10) {
			sprintf(&str[x], "%ld", o[i]);
			if (o[i] >= 100)
				x++;
			x += 2;
		} else {
			str[x++] = o[i] + '0';
		}
	}
	str[x] = 0;
	return str;
}

static char *value2str(netsnmp_variable_list *v, char *buf, size_t len)
{
	snprint_value(buf, len, v->name, v->name_length, v);
	return buf;
}

static int triggers_threshold(traf_info *val, traf_info *thresh)
{
	if (!thresh->pct && !thresh->bps)
		return 0;

	if (thresh->use_pct && val->pct >= thresh->pct)
		return 1;
	if (val->bps >= thresh->bps)
		return 1;

	return 0;
}

/* calculates the microsecond delta of a timeval and returns it as an */
static u64 tv_delta_usec(struct timeval *then, struct timeval *now)
{
	u64 usec, seconds;

	/* overflow protection */
	if (then->tv_sec > now->tv_sec)
		return 0;

	seconds = now->tv_sec - then->tv_sec;
	usec = (now->tv_usec - then->tv_usec);
	if (now->tv_usec < then->tv_usec) {
		seconds--;
		usec += 1000000;
	}

	if (seconds)
		usec += seconds * 1000000;

	return usec;
}

static void parse_if_errors(traf_info *in, traf_info *out, if_info *old, if_info *ii, int options)
{
	double seconds;
	u64 usecs;

	memset(in, 0, sizeof(*in));
	memset(out, 0, sizeof(*out));
	usecs = tv_delta_usec(&old->tv, &ii->tv);
	seconds = usecs / 1000000.0;
	seconds = ii->tv.tv_sec - old->tv.tv_sec;
	if (!seconds)
		seconds = 0.5;

	in->errors = ii->InErrors - old->InErrors;
	out->errors = ii->OutErrors - old->OutErrors;
	if (!(options & IFINFO_ERRORS_IGNORE_DISCARDS)) {
		in->discards = ii->InDiscards - old->InDiscards;
		out->discards = ii->OutDiscards - old->OutDiscards;
	}
	in->bps = (in->errors + in->discards) / (double)seconds;
	out->bps = (out->errors + out->discards) / (double)seconds;
}

static void parse_if_pps(traf_info *in, traf_info *out, if_info *old, if_info *ii)
{
	double seconds;
	u64 usecs;

	usecs = tv_delta_usec(&old->tv, &ii->tv);
	seconds = usecs / 1000000;
	if (!(seconds = ii->tv.tv_sec - old->tv.tv_sec))
		seconds = 0.5;

	memset(in, 0, sizeof(*in));
	memset(out, 0, sizeof(*out));
	in->ucastpps = (ii->InUcastPkts - old->InUcastPkts) / (double)seconds;
	out->ucastpps = (ii->OutUcastPkts - old->OutUcastPkts) / (double)seconds;
	in->bcastpps = (ii->InBroadcastPkts - old->InBroadcastPkts) / (double)seconds;
	out->bcastpps = (ii->OutBroadcastPkts - old->OutBroadcastPkts) / (double)seconds;
	in->mcastpps = (ii->InMulticastPkts - old->InMulticastPkts) / (double)seconds;
	out->mcastpps = (ii->OutMulticastPkts - old->OutMulticastPkts) / (double)seconds;
}

static int walkback(netsnmp_session *ss, netsnmp_variable_list *vars, void *arg, void *arg2)
{
	int i = 0;
	char *str = (char *)arg;
	int *index = (int *)arg2;

	netsnmp_variable_list *v;

	if (!str || !index)
		return 0;

	for (v = vars; v; v = v->next_variable) {
		if (!v->val.string)
			continue;
		v->val.string[v->val_len] = 0;
		if (!strcmp((char *)str, (char *)v->val.string)) {
			*index = v->name[v->name_length - 1];
			return SNMPWALK_STOP;
		}
		i++;
	}

	return 0;
}

static int snmpwalk(netsnmp_session *ss, const char *base_oid, walker_callback(func), void *arg, void *arg2)
{
	netsnmp_pdu	*pdu, *response;
	netsnmp_variable_list *vars;
	oid name[MAX_OID_LEN];
	size_t name_length;
	oid root[MAX_OID_LEN];
	size_t rootlen;
	oid end_oid[MAX_OID_LEN];
	size_t end_len = 0;
	int count, running, status = STAT_ERROR, exitval = 0;
	int result;

	/*
	 * get the initial object and subtree
	 */
	rootlen = MAX_OID_LEN;
	if (snmp_parse_oid(base_oid, root, &rootlen) == NULL) {
		printf("UNKNOWN - Failed to add %s as root for snmp traversal: %s\n",
			   base_oid, snmp_api_errstring(snmp_errno));
		exit(STATE_UNKNOWN);
	}

	memmove(end_oid, root, rootlen*sizeof(oid));
	end_len = rootlen;
	end_oid[end_len-1]++;

	/*
	 * get first object to start walk
	 */
	memmove(name, root, rootlen * sizeof(oid));
	name_length = rootlen;

	running = 1;
	while (running) {
		/*
		 * create PDU for GETNEXT request and add object name to request
		 */
		pdu = snmp_pdu_create(SNMP_MSG_GETNEXT);
		snmp_add_null_var(pdu, name, name_length);

		/* do the request */
		status = snmp_synch_response(ss, pdu, &response);
		if (status != STAT_SUCCESS) {
			if (status == STAT_TIMEOUT) {
				printf("Timeout: No Response from %s\n", ss->peername);
			}
			else {
				/* status == STAT_ERROR */
				printf("SNMP error when querying %s\n", ss->peername);
			}
			running = 0;
			exitval = 1;
			break;
		}

		if (response->errstat != SNMP_ERR_NOERROR) {
			/* error in response, print it */
			fprintf(stderr, "Error in packet.\nReason: %s\n",
					snmp_errstring(response->errstat));
			if (response->errindex != 0) {
				fprintf(stderr, "Failed object: ");
				for (count = 1, vars = response->variables;
					 vars && count != response->errindex;
					 vars = vars->next_variable, count++)
					/*EMPTY*/;
				if (vars)
					fprint_objid(stderr, vars->name, vars->name_length);
				fprintf(stderr, "\n");
			}
			exitval = 2;
			break;
		}

		/* check resulting variables */
		for (vars = response->variables; vars;
			 vars = vars->next_variable) {
			if (snmp_oid_compare(end_oid, end_len, vars->name, vars->name_length) <= 0) {
				/* not part of this subtree */
				running = 0;
				continue;
			}

			/* found a proper variable, so handle it it */
			result = func(ss, vars, arg, arg2);
			if (result == SNMPWALK_STOP) {
				running = 0;
				break;
			}

			if ((vars->type == SNMP_ENDOFMIBVIEW) ||
				(vars->type == SNMP_NOSUCHOBJECT) ||
				(vars->type == SNMP_NOSUCHINSTANCE))
			{
				running = 0;
				break;
			}

			memmove((char *) name, (char *) vars->name,
					vars->name_length * sizeof(oid));
			name_length = vars->name_length;
		}
		if (response)
			snmp_free_pdu(response);
	}

	return exitval;
}

static int fetch_ifkey(netsnmp_session *ss, const char *desc, const char *name)
{
	int key = -1;

	if (name && !desc) {
		snmpwalk(ss, ".1.3.6.1.2.1.31.1.1.1.1", walkback, (void *)name, &key);
	} else {
		snmpwalk(ss, ".1.3.6.1.2.1.2.2.1.2", walkback, (void *)desc, &key);
	}

	return key;
}

/* oid's are 1-indexed, while we use 0-indexing */
#define FLAGIFY_OID(i) (1 << ((i) - 1))
#define IFINFO_TRAFFIC 1
#define IFINFO_ERRORS 2
#define IFINFO_STATUS 3
#define IFINFO_PPS 4
#define IFINFO_ALL 5

#define IFTABLE_PREFIX ".1.3.6.1.2.1.2.2.1"
#define IFXTABLE_PREFIX ".1.3.6.1.2.1.31.1.1.1"

/* camelcase here, since we follow the MIB naming conventions */
#define IFOID_Index             1
#define IFOID_Descr             2
#define IFOID_Type              3
#define IFOID_Mtu               4
#define IFOID_Speed             5
#define IFOID_PhysAddress       6
#define IFOID_AdminStatus       7
#define IFOID_OperStatus        8
#define IFOID_LastChange        9
#define IFOID_InOctets         10
#define IFOID_InUcastPkts      11
#define IFOID_InNUcastPkts     12
#define IFOID_InDiscards       13
#define IFOID_InErrors         14
#define IFOID_InUnknownProtos  15
#define IFOID_OutOctets        16
#define IFOID_OutUcastPkts     17
#define IFOID_OutNUcastPkts    18
#define IFOID_OutDiscards      19
#define IFOID_OutErrors        20
#define IFOID_OutQLen          21
#define IFOID_Specific         22

/* in .1.3.6.1.2.1.31 subtree (the ifXTable) */
#define IFOID_Name                        1
#define IFOID_InMulticastPkts             2
#define IFOID_InBroadcastPkts             3
#define IFOID_OutMulticastPkts            4
#define IFOID_OutBroadcastPkts            5
#define IFOID_HCInOctets                  6
#define IFOID_HCInUcastPkts               7
#define IFOID_HCInMulticastPkts           8
#define IFOID_HCInBroadcastPkts           9
#define IFOID_HCOutOctets                10
#define IFOID_HCOutUcastPkts             11
#define IFOID_HCOutMulticastPkts         12
#define IFOID_HCOutBroadcastPkts         13
#define IFOID_LinkUpDownTrapEnable       14
#define IFOID_HighSpeed                  15
#define IFOID_PromiscuousMode            16
#define IFOID_ConnectorPresent           17
#define IFOID_Alias                      18
#define IFOID_CounterDiscontinuityTime   19

#define IFERRORS_REQUIRED (0 \
	| FLAGIFY_OID(IFOID_InDiscards) | FLAGIFY_OID(IFOID_InErrors) \
	| FLAGIFY_OID(IFOID_OutDiscards) | FLAGIFY_OID(IFOID_OutErrors) \
)
#define IFERRORS_WANTED IFERRORS_REQUIRED

#define IFSTATUS_REQUIRED (FLAGIFY_OID(IFOID_AdminStatus) | (FLAGIFY_OID(IFOID_OperStatus)))
#define IFPPS_REQUIRED (FLAGIFY_OID(IFOID_InUcastPkts) | (FLAGIFY_OID(IFOID_InNUcastPkts)))
#define IFPPS_WANTED IFPPS_REQUIRED
#define IFPPS_XREQUIRED ( 0 \
	| FLAGIFY_OID(IFOID_HCInUcastPkts) | FLAGIFY_OID(IFOID_HCOutUcastPkts) \
	| FLAGIFY_OID(IFOID_HCInMulticastPkts) | FLAGIFY_OID(IFOID_HCOutMulticastPkts) \
	| FLAGIFY_OID(IFOID_HCInBroadcastPkts) | FLAGIFY_OID(IFOID_HCOutBroadcastPkts) \
)
#define IFPPS_XWANTED IFPPS_XREQUIRED


/* either of these will work just fine */
#define IFTRAFFIC_REQUIRED (FLAGIFY_OID(IFOID_InOctets) | FLAGIFY_OID(IFOID_OutOctets))
#define IFTRAFFIC_XREQUIRED (FLAGIFY_OID(IFOID_HCInOctets) | FLAGIFY_OID(IFOID_HCOutOctets))
#define IFTRAFFIC_WANTED (IFTRAFFIC_REQUIRED | FLAGIFY_OID(IFOID_Speed))
#define IFTRAFFIC_XWANTED (IFTRAFFIC_XREQUIRED | FLAGIFY_OID(IFOID_HighSpeed))

#define ifoid_handle_int(n) \
case IFOID_##n: ii->n = *v->val.integer & 0xffffffff; break

#define ifoid_handle_c32(n) ifoid_handle_int(n)

#define ifoid_handle_c64(Prefix, n) \
case IFOID_##Prefix##n: ii->n = ((u_int64_t) v->val.counter64->high) << 32 | v->val.counter64->low; break


#define snmpif_name_case(x) case IFOID_##x: return #x
static const char *snmpif_oidname(int code)
{
	switch (code) {
	 snmpif_name_case(Index);
	 snmpif_name_case(Descr);
	 snmpif_name_case(Type);
	 snmpif_name_case(Mtu);
	 snmpif_name_case(Speed);
	 snmpif_name_case(PhysAddress);
	 snmpif_name_case(AdminStatus);
	 snmpif_name_case(OperStatus);
	 snmpif_name_case(LastChange);
	 snmpif_name_case(InOctets);
	 snmpif_name_case(InUcastPkts);
	 snmpif_name_case(InNUcastPkts);
	 snmpif_name_case(InDiscards);
	 snmpif_name_case(InErrors);
	 snmpif_name_case(InUnknownProtos);
	 snmpif_name_case(OutOctets);
	 snmpif_name_case(OutUcastPkts);
	 snmpif_name_case(OutNUcastPkts);
	 snmpif_name_case(OutQLen);
	 snmpif_name_case(Specific);
	 default:
		return "unknown";
	}
}

static const char *snmpifx_oidname(int code)
{
	switch (code) {
	 snmpif_name_case(Name);
	 snmpif_name_case(InMulticastPkts);
	 snmpif_name_case(InBroadcastPkts);
	 snmpif_name_case(OutMulticastPkts);
	 snmpif_name_case(OutBroadcastPkts);
	 snmpif_name_case(HCInOctets);
	 snmpif_name_case(HCInUcastPkts);
	 snmpif_name_case(HCInMulticastPkts);
	 snmpif_name_case(HCInBroadcastPkts);
	 snmpif_name_case(HCOutOctets);
	 snmpif_name_case(HCOutUcastPkts);
	 snmpif_name_case(HCOutMulticastPkts);
	 snmpif_name_case(HCOutBroadcastPkts);
	 snmpif_name_case(LinkUpDownTrapEnable);
	 snmpif_name_case(HighSpeed);
	 snmpif_name_case(PromiscuousMode);
	 snmpif_name_case(ConnectorPresent);
	 snmpif_name_case(Alias);
	 snmpif_name_case(CounterDiscontinuityTime);
	 default:
		return "unknown";
	}
}

/* some silly helpers */
static int ifstatus_status(int status, int dormant_state)
{
	switch (status) {
	case IFSTATUS_UP: return STATE_OK;
	case IFSTATUS_DOWN: return STATE_CRITICAL;
	case IFSTATUS_TESTING: return STATE_WARNING;
	case IFSTATUS_UNKNOWN: return STATE_WARNING;
	case IFSTATUS_DORMANT: return dormant_state;
	case IFSTATUS_NOTPRESENT: return STATE_CRITICAL;
	case IFSTATUS_LOWERLAYERDOWN: return STATE_CRITICAL;
	}
	return STATE_UNKNOWN;
}

static const char *ifstatus_str(int status)
{
	switch (status) {
	case IFSTATUS_UP: return "up";
	case IFSTATUS_DOWN: return "down";
	case IFSTATUS_TESTING: /* fallthrough for backwards compatibility */
	case IFSTATUS_UNKNOWN: return "down (testing/unknown)";
	case IFSTATUS_DORMANT: return "dormant";
	case IFSTATUS_NOTPRESENT:
		return "notPresent - possible hotswap in progress";
	case IFSTATUS_LOWERLAYERDOWN:
		return "down due to lower layer being down";
	}

	return "not available on this host";
}

/*
 * This function takes an snmp-subtree and walks all subtrees
 * of that subtree that are marked in the 'mask' argument, using
 * 'key' as the final entry in the oid to reach the leaf node.
 *
 * if base_oid == .1.3.6.1 and mask == 5 and key = 5, we would
 * add .1.3.6.1.0.5 and .1.3.6.1.2.5 to be fetched by the next
 * request using *pdu.
 */
static int pdu_add_keyed_subtree(netsnmp_pdu *pdu, const char *base_oid, int mask, int key)
{
	oid o[MAX_OID_LEN];
	size_t len = MAX_OID_LEN;
	int i = 0;

	if (key < 0)
		return key;

	memset(o, 0, sizeof(o));
	snmp_parse_oid(base_oid, o, &len);
	len++;
	o[len++] = key;
	/* snmp trees are 1-indexed, so i starts at one */
	for (i = 1; mask; i++, mask >>= 1) {
		if (!(mask & 1))
			continue;

		o[len - 2] = i;
		snmp_add_null_var(pdu, o, len);
	}
	return 0;
}

int handle_snmp_if_var(struct if_info *ii, netsnmp_variable_list *v)
{
	int ioid, xtable;

	/* next-to-last is the interesting one */
	ioid = v->name[v->name_length - 2];

	xtable = v->name[6] == 31;
	if (v->type == SNMP_ENDOFMIBVIEW ||
		v->type == SNMP_NOSUCHOBJECT ||
		v->type == SNMP_NOSUCHINSTANCE ||
		v->type == ASN_NULL)
	{
		debug("Early return (v->type is %d: %s)\n", v->type,
			  v->type == SNMP_ENDOFMIBVIEW ? "ENDOFMIBVIEW" :
			  v->type == SNMP_NOSUCHOBJECT ? "NOSUCHOBJECT" :
			  v->type == SNMP_NOSUCHINSTANCE ? "NOSUCHINSTANCE" :
			  v->type == ASN_NULL ? "ASN_NULL" : "Can't touch this. Hammertime!");

		/*
		 * not all devices support the ifXTable, so try to be graceful
		 * when we encounter one that doesn't. This should work well,
		 * since we don't actually *need* those variables to do the
		 * job (and yes, that's a hack. Live with it)
		 */
		if (xtable && (v->type == ASN_NULL || v->type == SNMP_NOSUCHOBJECT)) {
			return 0;
		}
		return -v->type;
	}
	if (v->name_length == uptime_oid_len && !memcmp(uptime_oid, v->name, uptime_oid_len)) {
		device_uptime = *v->val.integer;
		debug("    Device uptime: %lu\n", device_uptime);
		return 0;
	}

	debug("    xtable is %d. Type is %2d (%s)\n", xtable, ioid,
		xtable ? snmpifx_oidname(ioid) : snmpif_oidname(ioid));


	/*
	 * second to last digit in response OID
	 * determines where we put it, since some daemons
	 * will deliver the response in random order
	 */
	if (!xtable) {
		switch (ioid) {
			ifoid_handle_int(Speed);
			ifoid_handle_int(Type);
			ifoid_handle_int(AdminStatus);
			ifoid_handle_int(OperStatus);
			ifoid_handle_c32(InOctets);
			ifoid_handle_c32(InDiscards);
			ifoid_handle_c32(InErrors);
			ifoid_handle_c32(OutOctets);
			ifoid_handle_c32(OutDiscards);
			ifoid_handle_c32(OutErrors);
			ifoid_handle_c32(InUcastPkts);
			ifoid_handle_c32(OutUcastPkts);
		case IFOID_Descr:
			ii->Descr = strndup((char *)v->val.string, v->val_len);
			break;
		default:
			debug("Unhandled ifTable variable %s type: %d\n",
				oid2str(v->name, v->name_length), v->type);
		}
	} else {
		unsigned long c64;

		switch (ioid) {
			ifoid_handle_int(HighSpeed);

			/*
			 * some equipment (mainly Cisco) has serial interfaces
			 * where the HC*Octets counters are zero even though
			 * they're supported. Since we prefer those even if we
			 * get both in the request, they tend to overwrite the
			 * 32-bit octet counters.
			 */
		case IFOID_HCOutOctets:
			c64 = ((u_int64_t) v->val.counter64->high) << 32 | v->val.counter64->low;
			if (c64 || !ii->OutOctets)
				ii->OutOctets = c64;
			break;

		case IFOID_HCInOctets:
			c64 = ((u_int64_t) v->val.counter64->high) << 32 | v->val.counter64->low;
			if (c64 || !ii->InOctets)
				ii->InOctets = c64;
			break;

		ifoid_handle_c64(HC, InUcastPkts);
		ifoid_handle_c64(HC, OutUcastPkts);
		ifoid_handle_c64(, InBroadcastPkts);
		ifoid_handle_c64(HC, InBroadcastPkts);
		ifoid_handle_c64(, OutBroadcastPkts);
		ifoid_handle_c64(HC, OutBroadcastPkts);
		ifoid_handle_c64(, InMulticastPkts);
		ifoid_handle_c64(HC, InMulticastPkts);
		ifoid_handle_c64(, OutMulticastPkts);
		ifoid_handle_c64(HC, OutMulticastPkts);
		case IFOID_Name:
			ii->Name = strndup((char *)v->val.string, v->val_len);
			break;
		default:
			debug("Unhandled ifXEntry variable %s type: %d\n",
			      oid2str(v->name, v->name_length), v->type);
		}
	}

	if (xtable) {
		ii->xvars |= FLAGIFY_OID(ioid);
	} else {
		ii->vars |= FLAGIFY_OID(ioid);
	}
	ii->valid_vars++;
	return 0;
}

struct interface_table {
	struct if_info *list;
	unsigned int interfaces;
	unsigned int alloc;   /* space allocated for this list */
	unsigned int idx;
};

static int if_info_idxcmp(const void *a_, const void *b_)
{
	const struct if_info *a = (struct if_info *)a_;
	const struct if_info *b = (struct if_info *)b_;

	if (a->Index == b->Index)
		return 0;

	/* handle 'Index not set' specially */
	if (!a->Index)
		return 1;
	if (!b->Index)
		return -1;

	return a->Index - b->Index;
}

static int walk_types(netsnmp_session *ss, netsnmp_variable_list *v, void *table_, void *discard)
{
	struct interface_table *table = (struct interface_table *)table_;
	unsigned int idx, val;

	val = *v->val.integer;
	if (val == 24) {
		/* software loopback interface */
		return 0;
	}

	idx = v->name[v->name_length - 1];
	table->list[table->interfaces].Index = idx;
	table->list[table->interfaces].Type = val;
	table->interfaces++;

	return 0;
}

static void print_if_info(const char *name, if_info *ii)
{
	debug("Dumping if_info '%s'\n", name);
	debug("  AdminStatus: %d\n", ii->AdminStatus);
	debug("  Descr:       %s\n", ii->Descr);
	debug("  HighSpeed:   %lu\n", ii->HighSpeed);
	debug("  InDiscards:  %u\n", ii->InDiscards);
	debug("  InErrors:    %u\n", ii->InErrors);
	debug("  InOctets:    %lu\n", ii->InOctets);
	debug("  Index:       %d\n", ii->Index);
	debug("  Name:        %s\n", ii->Name);
	debug("  OperStatus:  %d\n", ii->OperStatus);
	debug("  OutDiscards: %u\n", ii->OutDiscards);
	debug("  OutErrors:   %u\n", ii->OutErrors);
	debug("  OutOctets:   %lu\n", ii->OutOctets);
	debug("  Speed:       %lu\n", ii->Speed);
	debug("  Type:        %u (%s)\n", ii->Type, iftype_str(ii->Type));
	debug("  InBroadcastpkts:    %u\n", ii->InBroadcastPkts);
	debug("  OutBroadcastpkts:   %u\n", ii->OutBroadcastPkts);
	debug("  InUcastpkts:        %u\n", ii->InUcastPkts);
	debug("  OutUcastpkts:       %u\n", ii->OutUcastPkts);
	debug("  InMulticastpkts:    %u\n", ii->InMulticastPkts);
	debug("  OutMulticastpkts:   %u\n", ii->OutMulticastPkts);
}

static int fetch_if_info(netsnmp_session *ss, int key, if_info *ii, int type)
{
	netsnmp_pdu *pdu, *response = NULL;
	int status, oidkeys = 0, xkeys = 0;
	netsnmp_variable_list *v;

	if (key < 0)
		return key;

	ii->Index = key;

	pdu = snmp_pdu_create(SNMP_MSG_GET);

	/* IFOID_Name is in the IfXTable */
	xkeys |= FLAGIFY_OID(IFOID_Name);
	oidkeys = FLAGIFY_OID(IFOID_Type);

	if (!ii->Descr && (0 || type == IFINFO_STATUS)) {
		oidkeys |= FLAGIFY_OID(IFOID_Descr);
	}
	oidkeys |= FLAGIFY_OID(IFOID_Descr);

	if (type == IFINFO_STATUS || type == IFINFO_ALL) {
		oidkeys |= IFSTATUS_REQUIRED;
	}
	if (type == IFINFO_ERRORS || type == IFINFO_ALL) {
		oidkeys |= IFERRORS_WANTED;
	}
	/* IFINFO_ALL doesn't do traffic yet */
	if (type == IFINFO_TRAFFIC) {
		/*
		 * for traffic, we fetch uptime so we can do accurate
		 * calculations if the device was reset since we last
		 * checked it
		 */
		snmp_parse_oid(".1.3.6.1.2.1.1.3.0", uptime_oid, &uptime_oid_len);
		snmp_add_null_var(pdu, uptime_oid, uptime_oid_len);

		oidkeys |= IFTRAFFIC_WANTED;
		xkeys |= IFTRAFFIC_XWANTED;
	}

	if (type == IFINFO_PPS) {
		oidkeys |= IFPPS_WANTED;
		xkeys |= IFPPS_XWANTED;
	}

	pdu_add_keyed_subtree(pdu, IFTABLE_PREFIX, oidkeys, key);
	if (!ignore_ifx) {
		pdu_add_keyed_subtree(pdu, IFXTABLE_PREFIX, xkeys, key);
	}

	debug("##### FETCH REQUEST #####\n");
	for (v = pdu->variables; v; v = v->next_variable) {
		debug("  REQUEST %s\n", oid2str(v->name, v->name_length));
	}
	debug("### FETCH REQUEST END ###\n");

	status = snmp_synch_response(ss, pdu, &response);
	if (status != STAT_SUCCESS) {
		printf("UNKNOWN - SNMP query error: %s\n",
			   snmp_api_errstring(snmp_errno));
		exit(STATE_UNKNOWN);
	}

	for (v = response->variables; v; v = v->next_variable) {
		char buf[1024];
		debug("  RESPONSE %s = %s\n", oid2str(v->name, v->name_length), value2str(v, buf, sizeof(buf)));
		handle_snmp_if_var(ii, v);
	}
	/* timestamp mark it for microsecond precision */
	gettimeofday(&ii->tv, NULL);
	snmp_free_pdu(response);
	print_if_info("newly fetched", ii);

	return 0;
}

struct interface_table *load_if_table(const char *path)
{
	struct stat st;
	int ret;
	static struct interface_table t;

	if (stat(path, &st) < 0) {
		printf("Failed to stat() %s: %m\n", path);
		return NULL;
	}

	t.alloc = st.st_size / sizeof(struct if_info);
	t.list = malloc(st.st_size);
	ret = db_load(path, t.list, st.st_size);
	if (ret < 0) {
		return NULL;
		printf("Failed to load interface table from '%s'\n", path);
	}

	return &t;
}

static void parse_key(netsnmp_session *ss, char *arg, char *pass, u_char *key, size_t *len)
{
	if (*pass == '0' && pass[1] == 'x') {
		if (!snmp_hex_to_binary((u_char **)&pass, len, (size_t *)key, 0, pass)) {
			usage("Bad key value for %s\n", arg);
		}
	} else {
		if (generate_Ku(ss->securityAuthProto,
						ss->securityAuthProtoLen,
						(u_char *)pass, strlen(pass),
						key, len) != SNMPERR_SUCCESS)
		{
			usage("Error generating Ku from authentication password '%s'\n", pass);
		}
	}
}

static double parse_ifspeed(const char *str)
{
	char *endp;
	double speed;

	errno = 0;
	speed = strtod(str, &endp);
	if (errno == ERANGE) {
		printf("%s is too large to handle properly\n", str);
		exit(STATE_UNKNOWN);
	}
	if (endp && *endp) {
		if (*endp == 'M' || *endp == 'm') {
			speed *= 1000000;
		} else if (*endp == 'G' || *endp == 'g') {
			speed *= 1000000000;
		} else if (*endp == 'K' || *endp == 'k') {
			speed *= 1000;
		} else if (*endp == 'T' || *endp == 't') {
			speed *= 1000000000000;
		}
	}
	return speed;
}

static int operstatus_format(int state, if_info *ii, const char *msg)
{
	printf("%s: Interface %s (index %d) is %s.\n",
		   state_str(state), interface_name(ii), ii->Index, msg);

	return state;
}

static int parse_threshold(const char *str, u64 speed, traf_info *ti, int pref)
{
	char *endp;
	int determinator;

	if (!ti)
		return 0;

	memset(ti, 0, sizeof(*ti));

	if (!str)
		return 0;

	determinator = str[strlen(str) - 1];
	if (isdigit(determinator)) {
		determinator = pref;
	}

	/* avoid divide-by-zero errors */
	if (!speed)
		speed = 1;

	if (determinator == '%') {
		errno = 0;
		ti->pct = strtod(str, &endp);
		ti->bps = ti->pct * speed / 100;
		ti->use_pct = 1;
	} else {
		ti->bps = parse_ifspeed(str);
		/* avoid rounding errors */
		ti->pct = (ti->bps * 100.0) / speed;
	}

	return ti->use_pct;
}


#define WRAP32 0xffffffffUL

static int parse_if_traffic(traf_info *in, traf_info *out, if_info *old, if_info *ii, u64 ifspeed)
{
	u64 usec, seconds;
	if_info real;
	int wrapped = 0;

	usec = tv_delta_usec(&old->tv, &ii->tv);
	seconds = usec / 1000000;

	if (!seconds)
		seconds = 1;

	if (!ifspeed) {
		ifspeed = compute_ifspeed(ii);
	}

	memset(&real, 0, sizeof(real));
	/*
	 * "Handle" corner-case 4211: Check used to be done via snmpv2,
	 * but is now done via snmpv1. WRAP32 is a-power-of-2-minus-1,
	 * so this is equivalent to a fast modulo
	 */
	if (!ii->HighSpeed && (old->InOctets > WRAP32 || old->OutOctets > WRAP32)) {
		old->InOctets &= WRAP32;
		old->OutOctets &= WRAP32;
	}
	/*
	 * corner-case 4212: Check used to be done via snmpv1 but is
	 * now done via snmpv2
	 */
	if (!old->HighSpeed && (ii->InOctets > WRAP32 || ii->OutOctets > WRAP32)) {
		real.InOctets = ii->InOctets;
		real.OutOctets = ii->OutOctets;
		ii->InOctets &= WRAP32;
		ii->OutOctets &= WRAP32;
	}
	/*
	 * If 32-bit counters wrapped around, we need to add some.
	 * If they wrapped around twice, we'll almost certainly be
	 * waaaay off the mark, but that can't be helped.
	 * We don't update the counters if we're running on 64-bit
	 * counters though, as it would take 58.5 years for them
	 * to wrap around on a fully saturated 10Gbps link.
	 */
	if (ii->InOctets < old->InOctets) {
		if (!ii->HighSpeed)
			ii->InOctets += WRAP32;
		wrapped = 1;
	}
	if (ii->OutOctets < old->OutOctets) {
		if (!ii->HighSpeed)
			ii->OutOctets += WRAP32;
		wrapped |= 2;
	}
#if DEBUG_WRAPPING == 1
	printf("    ii->InOctets: %lu\n", ii->InOctets);
	printf("   ii->OutOctets: %lu\n", ii->OutOctets);
	printf("   old->InOctets: %lu\n", old->InOctets);
	printf("  old->OutOctets: %lu\n", old->OutOctets);
#endif

	/*
	 * if the device was rebooted since we last checked it, we
	 * can get an accurate reading by treating last check as 0
	 * and occurring at the device start.
	 * We mustn't do that unless the counters were reset,
	 * however, since some devices have persistent storage
	 * that survive a reboot.
	 * We can't guarantee that that's what happened, as it's
	 * possible the counters wrapped around or were manually
	 * reset, but since both of those should be reasonably
	 * unusual, we go with this guess as a "best-effort" to
	 * get things right.
	 */
	if (wrapped) {
		if (device_uptime && device_uptime < seconds) {
			usec = 0;
			seconds = device_uptime;
			old->InOctets = 0;
			old->OutOctets = 0;
			wrapped = 0;
		} else if (ii->HighSpeed) {
			/*
			 * 64-bit counters will *never* wrap, so all we can do is
			 * just reset our own counters and continue monitoring
			 * properly on the next check
			 */
			printf("Counters reset or wrapped around. Database will be reinitialized\n");
			return STATE_UNKNOWN;
		}
	}

	/*
	 * multi-step math left since debugging retarded
	 * rounding errors. Just leave it as is...
	 */
	in->octets = ii->InOctets - old->InOctets;
	out->octets = ii->OutOctets - old->OutOctets;

	/* make sure we don't stash the incremented value */
	if (wrapped & 1 && !ii->HighSpeed) {
		ii->InOctets -= WRAP32;
	}
	if (wrapped & 2 && !ii->HighSpeed) {
		ii->OutOctets -= WRAP32;
	}

	in->ops = in->octets / (float)((float)usec / 1000000.0);
	out->ops = out->octets / (float)((float)usec / 1000000.0);
	in->bits = in->octets * 8;
	out->bits = out->octets * 8;
	in->bps = in->bits / (double)seconds;
	out->bps = out->bits / (double)seconds;

	/* preserve high precision values for next time around */
	if (real.InOctets || real.OutOctets) {
		ii->InOctets = real.InOctets;
		ii->OutOctets = real.OutOctets;
	}

	/* avoid rounding errors with some creative math */
	in->pct = ((in->bps * 1000) / (double)(ifspeed / 100)) / 1000;
	out->pct = ((out->bps * 1000) / (double)(ifspeed / 100)) / 1000;
	return 0;
}

int check_ifoperstatus(if_info *ii, struct options *o)
{
	const int s = ii->OperStatus;

	if (ii->AdminStatus == IFSTATUS_DOWN) {
		printf("%s: Interface %s (index %d) is administratively down\n",
			   state_str(o->admin_down_state), interface_name(ii), ii->Index);
		return o->admin_down_state;
	}

	return operstatus_format(ifstatus_status(s, o->dormant_state), ii, ifstatus_str(s));
}

static int check_traffic(if_info *old, if_info *ii,
			  threshold_info *warn, threshold_info *crit,
			  int options, u64 ifspeed)
{
	traf_info in, out;
	int s = STATE_OK;

	print_if_info("old", old);
	print_if_info("ii", ii);

	/*
	 * We can't trust our numbers at all if 64-bit counters
	 * wrapped around or were reset, so don't even try.
	 */
	if (ii->HighSpeed && (ii->InOctets < old->InOctets || ii->OutOctets < old->OutOctets)) {
		printf("Counter wraparound/reset. Database will be reinitialized\n");
		return STATE_OK;
	}

	/* in case there are errors in parse_if_traffic(), we need to return */
	if ((s = parse_if_traffic(&in, &out, old, ii, ifspeed)))
		return s;

	if (triggers_threshold(&in, &crit->in) || triggers_threshold(&out, &crit->out))
		s = STATE_CRITICAL;
	else if (triggers_threshold(&in, &warn->in) || triggers_threshold(&out, &warn->out))
		s = STATE_WARNING;

	printf("%s: Avg Traffic: %s (%.2lf%% / %s) in, %s (%.2lf%% / %s) out|",
		   state_str(s),
		   humanize_bits(in.bps), in.pct, humanize_bits(ifspeed),
		   humanize_bits(out.bps), out.pct, humanize_bits(ifspeed));

	if (options & IFINFO_TRAFFIC_BPS_PERFDATA) {
		printf("in_traffic=%.0lfbps;%.0lf;%.0lf;0;%llu out_traffic=%.0lfbps;%.0lf;%.0lf;0;%llu\n",
		       in.bps, warn->in.bps, crit->in.bps, ifspeed,
		       out.bps, warn->out.bps, crit->out.bps, ifspeed);
	} else if (options & IFINFO_TRAFFIC_NAGVIS_PERFDATA) {
		printf("inBandwidth=%.0lfbps;%.0lf;%.0lf;0;%llu outBandwidth=%.0lfbps;%.0lf;%.0lf;0;%llu ",
		       in.bps, warn->in.bps, crit->in.bps, ifspeed,
		       out.bps, warn->out.bps, crit->out.bps, ifspeed);
		printf("inUsage=%.2lf%%;%.2lf;%.2lf;; outUsage=%.2lf%%;%.2lf;%.2lf;;\n",
		       in.pct, warn->in.pct, crit->in.pct,
		       out.pct, warn->out.pct, crit->out.pct);
	} else {
		printf("in_traffic=%.2lf%%;%.2lf;%.2lf;; out_traffic=%.2lf%%;%.2lf;%.2lf;;\n",
		       in.pct, warn->in.pct, crit->in.pct,
		       out.pct, warn->out.pct, crit->out.pct);
	}

	return s;
}

static int check_iferrors(if_info *old, if_info *ii,
			   threshold_info *warn, threshold_info *crit,
			   int options)
{
	int s = STATE_OK;
	traf_info in, out;
	u64 usecs;
	double seconds;

	if (!old)
		return -1;

	usecs = tv_delta_usec(&old->tv, &ii->tv);
	seconds = usecs / 1000000.0;
	if (!seconds)
		seconds = 0.5;

	parse_if_errors(&in, &out, old, ii, options);
	memset(&in, 0, sizeof(in));
	memset(&out, 0, sizeof(out));
	in.errors = ii->InErrors - old->InErrors;
	out.errors = ii->OutErrors - old->OutErrors;
	if (!(options & IFINFO_ERRORS_IGNORE_DISCARDS)) {
		in.discards = ii->InDiscards - old->InDiscards;
		out.discards = ii->OutDiscards - old->OutDiscards;
	}

	/* here, bps = sum total of errors and discards */
	in.bps = (in.errors + in.discards) / (double)seconds;
	out.bps = (out.errors + out.discards) / (double)seconds;
	in.dps = (double)in.discards / (double)seconds;
	out.dps = (double)out.discards / (double)seconds;
	in.eps = (double)in.errors / (double)seconds;
	out.eps = (double)out.errors / (double)seconds;

	if (triggers_threshold(&in, &crit->in) || triggers_threshold(&out, &crit->out))
		s = STATE_CRITICAL;
	else if (triggers_threshold(&in, &warn->in) || triggers_threshold(&out, &warn->out))
		s = STATE_WARNING;

	printf("%s: Errors/sec = %.2lf :: IN - discards: %.2lf, errors: %.2lf :: ",
		  state_str(s), in.bps + out.bps, in.dps, in.eps);
	printf("OUT - discards: %.2lf, errors: %.2lf\n", out.dps, out.eps);
	printf("|discard_in=%.2lf;%.1lf;%.1lf;; discard_out=%.2lf;%.1lf;%.1lf;; ",
	       in.dps, warn->in.bps, crit->in.bps,
	       out.dps, warn->out.bps, crit->out.bps);
	printf("error_in=%.2lf;%.1lf;%.1lf;; error_out=%.2lf;%.1lf;%.1lf;;\n",
		   in.eps, warn->in.bps, crit->in.bps,
	       out.eps, warn->out.bps, crit->out.bps);

	return s;
}

static int check_ifpps(if_info *old, if_info *ii,
                     threshold_info *warn, threshold_info *crit,
                     int options)
{
	traf_info in, out;
	double in_pps, in_nupps, out_pps, out_nupps;
	double tot_pps;
	int s = STATE_OK;

	parse_if_pps(&in, &out, old, ii);
	in_nupps = in.bcastpps + in.mcastpps;
	in_pps = in_nupps + in.ucastpps;
	out_nupps = out.bcastpps + out.mcastpps;
	out_pps = out_nupps + out.ucastpps;
	tot_pps = in_pps + out_pps;
	if (triggers_threshold(&in, &crit->in) || triggers_threshold(&out, &crit->out))
		s = STATE_CRITICAL;
	else if (triggers_threshold(&in, &warn->in) || triggers_threshold(&out, &warn->out))
		s = STATE_WARNING;
	printf("%s: Packets/sec = %.2lf (in=%.2lf, out=%.2lf)"
	       " :: IN %.2lf - Uni: %.2lf, Broad: %.2lf, Multi: %.2lf, NonUni: %.2lf"
	       " :: OUT %.2lf - Uni: %.2lf, Broad: %.2lf, Multi: %.2lf, NonUni: %.2lf\n",
	       state_str(s), tot_pps, in_pps, out_pps,
	       in_pps, in.ucastpps, in.bcastpps, in.mcastpps, in_nupps,
	       out_pps, out.ucastpps, out.bcastpps, out.mcastpps, out_nupps);
	printf("| in_pps=%.2lfpps;%.2lf;%.2lf;; in_ucastpps=%.2lfpps;;;; "
	       " in_bcastpps=%.2lf;;;; in_mcastpps=%.2lf;;;; "
	       " out_pps=%.2lfpps;%.2lf;%.2lf;; out_ucastpps=%.2lfpps;;;; "
	       " out_bcastpps=%.2lf;;;; out_mcastpps=%.2lf;;;; "
	       " in_nonucastpps=%.2lfpps;;;; out_nonucastpps=%.2lfpps;;;;\n",
	       in_pps, warn->in.bps, crit->in.bps, in.ucastpps,
	       in.bcastpps, in.mcastpps,
	       out_pps, warn->out.bps, crit->out.bps, out.ucastpps,
	       out.bcastpps, out.mcastpps,
	       in_nupps, out_nupps);
	return s;
}

/*
 * we overalloc by lots, since we don't realloc().
 * In a customer network with 700k devices, the highest monitored
 * interface number is 2709, but we play it safe and assume switches
 * can use 16 bits to assign interface id.
 */
static int check_ifall(netsnmp_session *session, const char *path, const char *wstr, const char *cstr)
{
	netsnmp_session *ss, nsess;
	struct interface_table table, *old_table;
	unsigned int i, operstatus_bad = 0, errors_bad = 0, non_ignored = 0;
	u64 ifspeed = 1, last_speed = 0; /* can't be equal for first run */
	traf_info errin, errout;
	threshold_info warn, crit;
#if 0
	traf_info trafin, trafout;
#endif

	old_table = load_if_table(path);

	memset(&table, 0, sizeof(table));
	table.alloc = 8192;
	table.list = calloc(table.alloc, sizeof(*table.list));
	if (!table.list)
		die(STATE_UNKNOWN, "Failed to allocate memory\n");

	/*
	 * first walk we get ifType and ifIndex (since we get ifIndex
	 * from the OID), so we can ignore software loopback interfaces
	 */
	memcpy(&nsess, session, sizeof(nsess));
	ss = snmp_open(&nsess);
	snmpwalk(ss, ".1.3.6.1.2.1.2.2.1.3", walk_types, &table, NULL);
	snmp_close(ss);

	for (i = 0; i < table.interfaces; i++) {
		struct if_info *ii = &table.list[i];
		struct if_info *old;

		/* ignored due to wrong type, possibly */
		if (!ii->Index)
			continue;
		memcpy(&nsess, session, sizeof(nsess));
		ss = snmp_open(&nsess);
		fetch_if_info(ss, ii->Index, ii, IFINFO_ALL);
		snmp_close(ss);
		if (!ignore_ifx && !ii->valid_vars) {
			ignore_ifx = 1;
			memcpy(&nsess, session, sizeof(nsess));
			ss = snmp_open(&nsess);
			fetch_if_info(ss, ii->Index, ii, IFINFO_ALL);
			snmp_close(ss);
		}
#if 0
		debug("%u(%u); AStatus=%d (%s); OStatus=%d (%4s) InOctets=%lu OutOctets=%lu\n",
			   i, ii->Index,
			   ii->AdminStatus, ifstatus_str(ii->AdminStatus),
			   ii->OperStatus, ifstatus_str(ii->OperStatus),
			   ii->InOctets, ii->OutOctets);
#endif
		/* ignore turned off interfaces and those without traffic */
		if (ii->AdminStatus != IFSTATUS_UP) {
			continue;
		}
		if (!ii->InOctets && !ii->OutOctets) {
			continue;
		}
		/* if we couldn't get required variables, this could be a
		 * weird interface that doesn't track traffic (and maybe
		 * can't send or receive), so let's ignore it
		 */
		if ((ii->vars & IFTRAFFIC_REQUIRED) != IFTRAFFIC_REQUIRED &&
		    (ii->xvars & IFTRAFFIC_XREQUIRED) != IFTRAFFIC_XREQUIRED)
		{
			continue;
		}

		non_ignored++;

		if (ii->OperStatus != IFSTATUS_UP) {
			operstatus_bad++;
			/* don't measure errors on non-up interfaces */
			continue;
		}

		if (!old_table || old_table->interfaces >= old_table->alloc)
			continue;

		/*
		 * now try to get data from last run. We assume it's in sorted
		 * order, since all switches I've seen handle snmp properly in
		 * that respect
		 */
		old = bsearch(ii, old_table->list, old_table->alloc, sizeof(*ii), if_info_idxcmp);
		if (!old || old->Index != ii->Index) {
			continue;
		}

		/* ignore interfaces with no traffic at all since boot-time */
		if (!ii->InOctets && !ii->OutOctets)
			continue;

		ifspeed = compute_ifspeed(ii);
		if (ifspeed != last_speed) {
			last_speed = ifspeed;
			if (wstr) {
				parse_threshold(wstr, ifspeed, &warn.in, 'b');
				parse_threshold(wstr, ifspeed, &warn.out, 'b');
			}
			if (cstr) {
				parse_threshold(cstr, ifspeed, &crit.in, 'b');
				parse_threshold(cstr, ifspeed, &crit.out, 'b');
			}
		}
#if 0
		debug("old is set for if %u. whoptidoo\n", ii->Index);
		debug("Traffic delta in/out: %s / %s\n",
			   humanize_bytes(ii->InOctets - old->InOctets),
			   humanize_bytes(ii->OutOctets - old->OutOctets));

		debug("Error delta in/out: %u / %u\n",
			   ii->InErrors - old->InErrors,
			   ii->OutErrors - old->OutErrors);

		debug("%u; warn.bps: %.2lf; warn.pct: %.2lf; crit.bps: %.2lf; crit.pct: %.2lf\n",
			   ii->Index, warn.bps, warn.pct, crit.bps, crit.pct);
#endif
		parse_if_errors(&errin, &errout, old, ii, 0);
		if (wstr &&
			(triggers_threshold(&errin, &crit.in) ||
			 triggers_threshold(&errout, &crit.out) ||
			 triggers_threshold(&errin, &warn.in) ||
			 triggers_threshold(&errout, &warn.out)
			)
		   )
		{
			errors_bad++;
		}

#if 0
		/* no support for checking traffic yet */
		parse_if_traffic(&trafin, &trafout, old, ii, 0);
		if (triggers_threshold(&trafin, &crit) || triggers_threshold(&trafout, &crit))
			traffic_bad++;
		if (triggers_threshold(&trafin, &warn) || triggers_threshold(&trafout, &warn))
			traffic_bad++;
#endif
	}

	db_save(path, table.list, table.interfaces * sizeof(*table.list));

	if (!operstatus_bad && !errors_bad) {
		printf("OK: %u non-ignored interface%s working within thresholds\n",
			   non_ignored, non_ignored > 1 ? "s" : "");
		return STATE_OK;
	}

	printf("CRITICAL: ");
	if (operstatus_bad) {
		printf("%u trafficked interfaces with AdminStatus: UP and OperStatus: NOT UP.", operstatus_bad);
	}
	if (errors_bad) {
		printf("%s%u interfaces have too many errors/second.",
			   operstatus_bad ? "" : " ", errors_bad);
	}

	putchar('\n');

	return STATE_CRITICAL;
}

int grok_state_string(const char *s)
{
	switch (*s) {
	case 'i': case 'I': case 'o': case 'O': case '0':
		return STATE_OK;
	case 'w': case 'W': case '1':
		return STATE_WARNING;
	case 'c': case 'C': case '2':
		return STATE_CRITICAL;
	case 'u': case 'U': case '3':
		return STATE_UNKNOWN;
	}

	return -1;
}

#define prefixcmp(a, b) strncmp(a, b, strlen(b))
int check_snmpif(int argc, char ** argv)
{
	struct snmp_session session, nsess, *ss;
	int status = STATE_UNKNOWN, i = 0, key = 0;
	if_info ii;
	char *auth_pass = NULL, *auth_pass_arg = NULL;
	char *priv_pass = NULL, *priv_pass_arg = NULL;
	char *name = NULL, *descr = NULL;
	char *mode_str = NULL;
	int mode = 0;
	threshold_info warn, crit;
	char *wstr = NULL, *cstr = NULL;
	char *wstr_in = NULL, *cstr_in = NULL, *wstr_out = NULL, *cstr_out = NULL;
	int arg_start = 1;
	char *progname;
	char *db_dir = "/opt/monitor/var/snmpif";
	u64 ifspeed = 0;
	int options = 0, sampling_interval = 0;
	int preferred_threshold = '%';
	int enough_vars = 0;
	struct options opts;

	memset(&warn, 0, sizeof(warn));
	memset(&crit, 0, sizeof(crit));
	memset(&opts, 0, sizeof(opts));
	opts.dormant_state = STATE_CRITICAL;

	snmp_sess_init(&session);
	session.community = (u_char *)"public";
	session.community_len = (size_t)strlen((char *)session.community);
	session.peername = "localhost";
	session.version = SNMP_VERSION_1;
	session.timeout = 15 * 1000000;
	/*
	 * NOTE: We don't override the default retry value here, even though it
	 * occasionally leads to timeouts since doing so would cause problems
	 * with failing checks. The retry number seems to be necessary. It might be
	 * a *tad* too high in the general case, but we're better off just trusting
	 * net-snmp here and let users set the retries number themselves if they
	 * experience problems.
	 */

	if (argc < 2 ) {
		usage(NULL);
	}

	progname = basename(argv[0]);
	mode_str = argv[1];
	arg_start = 2;
	if (!strcmp(mode_str, "operstatus") || !strcmp(mode_str, "status")) {
		mode = IFINFO_STATUS;
	} else if (!strcmp(mode_str, "traffic")) {
		mode = IFINFO_TRAFFIC;
	} else if (!strcmp(mode_str, "errors")) {
		mode = IFINFO_ERRORS;
	} else if (!strcmp(mode_str, "pps")) {
		mode = IFINFO_PPS;
	} else if (!strcmp(mode_str, "all")) {
		mode = IFINFO_ALL;
	}

	if (!mode) {
		arg_start = 1;
		if (!strcmp(progname, "check_ifoperstatus") ||
			!strcmp(progname, "check_ifoperstatus.pl"))
		{
			mode = IFINFO_STATUS;
			mode_str = "operstatus";
		}
		else if (!strcmp(progname, "check_traffic") ||
				 !strcmp(progname, "check_traffic.pl"))
		{
			mode = IFINFO_TRAFFIC;
			mode_str = "traffic";
		}
		else if (!strcmp(progname, "check_iferrors") ||
				 !strcmp(progname, "check_iferrors.pl"))
		{
			mode = IFINFO_ERRORS;
			mode_str = "errors";
		}
		else if (!strcmp(progname, "check_ifpps") ||
		         !strcmp(progname, "check_ifpps.pl"))
		{
			mode = IFINFO_PPS;
			mode_str = "pps";
		}
		else if (!strcmp(progname, "check_snmpif_all") ||
				 !strcmp(progname, "check_if_all"))
		{
			mode = IFINFO_ALL;
			mode_str = "all";
		}
	}

	if (!mode) {
		usage("Mode of operation or executable name hack required\n");
	}

	switch (mode) {
	 case IFINFO_ERRORS: preferred_threshold = 'b'; break;
	 case IFINFO_PPS: preferred_threshold = 'p'; break;
	 default: preferred_threshold = '%';
	}

	for (i = arg_start; i < argc; i++) {
		char *arg = argv[i];
		char *opt = NULL;
		char *eq = strchr(arg, '=');

		if (eq) {
			*eq = 0;
			opt = eq + 1;
		} else if (i < argc - 1) {
			opt = argv[i + 1];
		}

		if (*arg != '-')
			usage("Unknown argument '%s'\n", arg);

		if (arg[1] == 'h' || !strcmp(arg, "--help"))
			usage(NULL);

		if (!strcmp(arg, "--debug")) {
			verbose++;
			continue;
		}

		/* check_iferrors specific */
		if (mode == IFINFO_ERRORS || mode == IFINFO_ALL) {
			if (arg[1] == 'O' || !strcmp(arg, "--only-errors")) {
				options |= IFINFO_ERRORS_IGNORE_DISCARDS;
				continue;
			}
		}

		/* check_traffic specific non-argument requiring options */
		if (mode == IFINFO_TRAFFIC) {
			if (arg[1] == 'B' || !strcmp(arg, "--perfdatabits")) {
				options |= IFINFO_TRAFFIC_BPS_PERFDATA;
				continue;
			}
			if (arg[1] == 'N' || !strcmp(arg, "--nagvis-perfdata")) {
				options |= IFINFO_TRAFFIC_NAGVIS_PERFDATA;
				continue;
			}
		}

		/* check_ifoperstatus specific */
		if (mode == IFINFO_STATUS) {
			if (arg[1] == 'D' || !strcmp(arg, "--admin-down")) {
				opts.admin_down_state = grok_state_string(opt);
				if (opts.admin_down_state < 0)
					usage("--admin-down requires a valid string (ok, warning, critical or unknown)\n");
				continue;
			}
		}

		if (!eq)
			i++;

		if (!opt) {
			usage("Illegal option or missing argument: %s\n", arg);
		}
		/* backwards compatibility only. We always hide community */
		if (!strcmp(arg, "--hide-community") || !strcmp(arg, "-hc")) {
			continue;
		}
		if (arg[1] == 'H' || !strcmp(arg, "--host")) {
			session.peername = opt;
			continue;
		}
		if (!prefixcmp(arg, "--warn-in")) {
			wstr_in = opt;
			continue;
		}
		if (!prefixcmp(arg, "--warn-out")) {
			wstr_out = opt;
			continue;
		}
		if (arg[1] == 'w' || !prefixcmp(arg, "--warn")) {
			wstr = wstr_in = wstr_out = opt;
			continue;
		}
		if (!prefixcmp(arg, "--crit-in")) {
			cstr_in = opt;
			continue;
		}
		if (!prefixcmp(arg, "--crit-out")) {
			cstr_out = opt;
			continue;
		}
		if (arg[1] == 'c' || !prefixcmp(arg, "--crit")) {
			cstr = cstr_in = cstr_out = opt;
			if (mode == IFINFO_STATUS) {
				usage("Unknown argument for mode 'status'\n");
			}
			continue;
		}
		if (arg[1] == 'C' || !strcmp(arg, "--community")) {
			session.community = (u_char *)opt;
			session.community_len = strlen(opt);
			continue;
		}
		if (arg[1] == 't' || !strcmp(arg, "--timeout")) {
			session.timeout = atoi(opt) * 1000000;
			continue;
		}
		if (arg[1] == 'r' || !strcmp(arg, "--retries")) {
			session.retries = atoi(opt);
			continue;
		}
		if (arg[1] == 'k' || !strcmp(arg, "--key") ||
			arg[1] == 'i' || !prefixcmp(arg, "--interface"))
		{
			if (!isdigit(*opt))
				usage("%s requires a numerical argument\n", arg);
			key = atoi(opt);
			if (key <= 0)
				usage("%s requires a positive index number\n", arg);
			continue;
		}
		if (arg[1] == 'd' || !strcmp(arg, "--description")) {
			descr = opt;
			continue;
		}
		if (arg[1] == 'n' || !strcmp(arg, "--name")) {
			name = opt;
			continue;
		}
		if (arg[1] == 'v' || !strcmp(arg, "--version") ||
			!strcmp(arg, "--snmp-version"))
		{
			if (*opt == '1')
				session.version = SNMP_VERSION_1;
			else if (*opt == '2')
				session.version = SNMP_VERSION_2c;
			else if (*opt == '3')
				session.version = SNMP_VERSION_3;
			else
				usage("%s requires 1, 2[c] or 3 as argument\n", arg);
			continue;
		}
		/* SNMP v3 crap goes here */
		if (arg[1] == 'L' || !strcmp(arg, "--seclevel")) {
			session.securityLevel = parse_secLevel_conf(arg, opt);
			if (session.securityLevel < 0) {
				usage("Illegal argument for option %s\n", arg);
			}
			continue;
		}
		if (arg[1] == 'U' || !strcmp(arg, "--secname")) {
			session.securityName = opt;
			session.securityNameLen = strlen(opt);
			continue;
		}
		if (arg[1] == 'a' || !prefixcmp(arg, "--authprot")) {
			if (*opt == 'm' || *opt == 'M') {
				session.securityAuthProto = usmHMACMD5AuthProtocol;
				session.securityAuthProtoLen = USM_AUTH_PROTO_MD5_LEN;
			} else if (*opt == 's' || *opt == 'S') {
				session.securityAuthProto = usmHMACSHA1AuthProtocol;
				session.securityAuthProtoLen = USM_AUTH_PROTO_SHA_LEN;
			} else {
				usage("%s requires 'md5' or 'sha1' as argument\n", arg);
			}
			continue;
		}
		if (arg[1] == 'A' || !strcmp(arg, "--authpass")) {
			auth_pass = opt;
			auth_pass_arg = arg;
			continue;
		}
		if (arg[1] == 'P' || !prefixcmp(arg, "--privprot")) {
			if (*opt == 'd' || *opt == 'D') {
				session.securityPrivProto = usmDESPrivProtocol;
				session.securityPrivProtoLen = USM_PRIV_PROTO_DES_LEN;
			} else if (*opt == 'a' || *opt == 'A') {
				session.securityPrivProto = usmAESPrivProtocol;
				session.securityPrivProtoLen = USM_PRIV_PROTO_AES_LEN;
			} else {
				usage("%s requires 'des' or 'aes' as argument\n", arg);
			}

			continue;
		}
		if (arg[1] == 'X' || !strcmp(arg, "--privpass")) {
			priv_pass = opt;
			priv_pass_arg = arg;
			continue;
		}

		/* 'traffic' and 'errors' specific */
		if (mode == IFINFO_TRAFFIC || mode == IFINFO_ERRORS) {
			if (arg[1] == 'S' || !strcmp(arg, "--sampling-interval")) {
				sampling_interval = atoi(opt);
				continue;
			}
		}

		/* check_traffic specific */
		if (mode == IFINFO_TRAFFIC) {
			if (arg[1] == 'b' || !strcmp(arg, "--bps")) {
				ifspeed = parse_ifspeed(opt);
				continue;
			}
		}
		usage("Unknown argument: %s\n", arg);
	}

	if (priv_pass || auth_pass || session.securityName
	    || session.securityPrivProtoLen || session.securityAuthProtoLen
	    || session.securityLevel)
	{
		session.version = SNMP_VERSION_3;
	}

	/*
	 * disable logging. net-snmp is a library and shouldn't log
	 * anything *anywhere* godsdammit! This is necessary since we
	 * trick it into not being a dumbass with the config junk
	 * below and it will otherwise complain (loudly) that it can't
	 * read the normal mibs, which we *really* don't want to read,
	 * since it completely kills performance.
	 */
	for (i = 0; i < LOG_DEBUG; i++) {
		(void)netsnmp_register_loghandler(NETSNMP_LOGHANDLER_NONE, i);
	}
	/* prevent init_snmp() from reading a bazillion ton of junk */
	netsnmp_set_mib_directory(":");
	netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_DISABLE_CONFIG_LOAD, 1);
	netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_PRINT_BARE_VALUE, 1);
	netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_QUICKE_PRINT, 1);
	/* SNMPv3 needs this for crypto setup */
	init_snmp("snmpif");

	if (wstr && mode == IFINFO_STATUS) {
		opts.dormant_state = grok_state_string(wstr);
		if (opts.dormant_state < 0) {
			usage("--warn requires a valid keyword (ok, warning, critical or unknown)\n");
		}
	}

	memset(&ii, 0, sizeof(ii));
	if (mode != IFINFO_ALL && !key && !descr && !name) {
		usage("One (and only one) of -[i|k], -d and -n must be supplied\n");
	}
	if (key && (descr || name)) {
		usage("Only one of -[i|k], -d and -n can be specified\n");
	}
	if (!session.peername) {
		usage("No hostname/address supplied\n");
	}

	if (auth_pass) {
		session.securityAuthKeyLen = sizeof(session.securityAuthKey);
		parse_key(&session, auth_pass_arg, auth_pass, session.securityAuthKey, &session.securityAuthKeyLen);
	}
	if (priv_pass) {
		session.securityPrivKeyLen = sizeof(session.securityPrivKey);
		if (session.securityPrivProto == NULL) {
#ifndef NETSNMP_DISABLE_DES
			session.securityPrivProto = snmp_duplicate_objid(usmDESPrivProtocol, USM_PRIV_PROTO_DES_LEN);
			session.securityPrivProtoLen = USM_PRIV_PROTO_DES_LEN;
#else
			session.securityPrivProto = snmp_duplicate_objid(usmAESPrivProtocol, USM_PRIV_PROTO_AES_LEN);
			session.securityPrivProtoLen = USM_PRIV_PROTO_AES_LEN;
#endif
		}
		parse_key(&session, priv_pass_arg, priv_pass, session.securityPrivKey, &session.securityPrivKeyLen);
	}

	if (mode == IFINFO_ALL) {
		const char *db_path = get_db_path(db_dir, session.peername, NULL, NULL);
		exit(check_ifall(&session, db_path, wstr, cstr));
	}

	/* 700 lines later we can finally issue a query */
	if (!key) {
		ss = snmp_open(&session);
		if (!ss) {
			printf("UNKNOWN - Failed to get snmp session handle\n");
			exit(STATE_UNKNOWN);
		}
		key = fetch_ifkey(ss, descr, name);
		snmp_close(ss);
		if (key < 0) {
			printf("Failed to get interface index for interface %s\n",
				   name ? name : descr);
			exit(STATE_UNKNOWN);
		}
	}

	ii.Index = key;
	ii.Descr = descr;
	ii.Name = name;
	memcpy(&nsess, &session, sizeof(nsess));
	ss = snmp_open(&nsess);
	if (!ss) {
		printf("UNKNOWN - Failed to get snmp session handle\n");
		exit(STATE_UNKNOWN);
	}
	fetch_if_info(ss, key, &ii, mode);
	snmp_close(ss);

	/*
	 * if this node gets cranky when we ask for IfXTable
	 * entries, we try again without requesting them and
	 * live with the poor precision from IfTable.
	 */
	if (!ignore_ifx && !ii.valid_vars) {
		ignore_ifx = 1;
		memcpy(&nsess, &session, sizeof(nsess));
		ss = snmp_open(&nsess);
		fetch_if_info(ss, key, &ii, mode);
		snmp_close(ss);
	}

	debug("Got %u valid variables for interface %s from %s\n",
		ii.valid_vars, interface_name(&ii), session.peername);
	if (!ii.valid_vars) {
		printf("UNKNOWN: No variables fetched for interface %s on %s. Bad auth credentials, wrong SNMP version or missing interface?\n",
		       interface_name(&ii), session.peername);

		exit(STATE_UNKNOWN);
	}

	/* make sure we read enough variables */
	if (mode == IFINFO_ERRORS && (ii.vars & IFERRORS_REQUIRED) == IFERRORS_REQUIRED)
		enough_vars = 1;
	if (mode == IFINFO_TRAFFIC) {
		if ((ii.vars & IFTRAFFIC_REQUIRED) == IFTRAFFIC_REQUIRED)
			enough_vars = 1;
		if ((ii.xvars & IFTRAFFIC_XREQUIRED) == IFTRAFFIC_XREQUIRED)
			enough_vars = 1;
	}
	if (mode == IFINFO_STATUS && (ii.vars & IFSTATUS_REQUIRED) == IFSTATUS_REQUIRED)
		enough_vars = 1;
	if (!enough_vars) {
		if (mode == IFINFO_ERRORS)
			printf("UNKNOWN: %s interface %d (%s) on %s doesn't support error or discard counters\n",
			       iftype_str(ii.Type), ii.Index, interface_name(&ii), session.peername);
		else if (mode == IFINFO_TRAFFIC)
			printf("UNKNOWN: %s interface %d (%s) on %s doesn't support traffic counters\n",
				   iftype_str(ii.Type), ii.Index, interface_name(&ii), session.peername);
		else
			printf("UNKNOWN: Not enough data returned from %s for %s interface %s (%u variables read).\n",
		           session.peername, iftype_str(ii.Type), interface_name(&ii), ii.valid_vars);
		exit(STATE_UNKNOWN);
	}

	if (mode == IFINFO_STATUS) {
		status = check_ifoperstatus(&ii, &opts);
	} else if (mode == IFINFO_TRAFFIC || mode == IFINFO_ERRORS || mode == IFINFO_PPS) {
		if_info old;
		const char *db_path = NULL;
		int save_ret = 0, load_ret = 0;

		if (mode == IFINFO_TRAFFIC && !ifspeed) {
			ifspeed = compute_ifspeed(&ii);
		}

		parse_threshold(wstr_in, ifspeed, &warn.in, preferred_threshold);
		parse_threshold(wstr_out, ifspeed, &warn.out, preferred_threshold);
		parse_threshold(cstr_in, ifspeed, &crit.in, preferred_threshold);
		parse_threshold(cstr_out, ifspeed, &crit.out, preferred_threshold);

		if (mode == IFINFO_ERRORS) {
			if (wstr || !warn.in.bps)
				warn.out.bps = warn.in.bps = 1.5;
			if (!cstr || !crit.in.bps)
				crit.out.bps = crit.in.bps = 2.5;
		}


		if (!sampling_interval) {
			db_path = get_db_path(db_dir, session.peername, mode_str, &ii);
			load_ret = load_if_data(db_path, &old);
			debug("Loaded old database from %s (ret=%d; %m)\n", db_path, load_ret);
		} else {
			memcpy(&old, &ii, sizeof(ii));
			memset(&ii, 0, sizeof(ii));
			free(old.Name);
			free(old.Descr);
			sleep(sampling_interval);
			ss = snmp_open(&session);
			fetch_if_info(ss, key, &ii, mode);
			snmp_close(ss);
		}

		if (!load_ret) {
			if (mode == IFINFO_TRAFFIC) {
				status = check_traffic(&old, &ii, &warn, &crit, options, ifspeed);
			} else if (mode == IFINFO_ERRORS) {
				status = check_iferrors(&old, &ii, &warn, &crit, options);
			} else if (mode == IFINFO_PPS) {
				status = check_ifpps(&old, &ii, &warn, &crit, options);
			}
		}

		if (!sampling_interval) {
			save_ret = save_if_data(db_path, &ii);
			if (load_ret) {
				if (!save_ret) {
					status = STATE_OK;
					printf("%s: Database initialized\n", state_str(status));
				} else {
					status = STATE_WARNING;
					printf("%s: Failed to stash current data in %s: %s\n",
						   state_str(status), db_path, strerror(-save_ret));
				}
			}
		}
	}

	free(ii.Name);
	free(ii.Descr);
	snmp_shutdown("snmpif");

	return status;
}
